/*
 * ElementAutomata.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 16-ott-03
 */
package xrel.analyzer;

import java.io.PrintStream;
import java.util.Vector;

import xrel.parser.SymElement;
import xrel.parser.SymElementKind;
import xrel.parser.SymTable;

/**
 * <p>Contains the methods to build and show the automata associated to 
 * the elements of the symbol table (types, patterns, expressions). </p>
 * <p>To build the automata simply call buildAll(). To show the
 * automata call dumpAutomata().</p>
 * <p>N.B.:
 * <ul>
 * <li>this is a static class so you will never need to allocate an 
 * object for this class</li>
 * <li>note that the word "element" here refers to elements in the 
 * symbol table and not to the XML elements, as it's the case of 
 * Hosoya's element automata</li>
 * </ul></p>
 * @author Fabrizio Bisi
 */
public class ElementAutomata implements SymElementKind {
	/** 
	 * <p>Writes the internal structure of the automata associated to 
	 * the elements of the symbol table. For each automaton it writes out 
	 * the set of the states, the subsets of initial and final states and
	 * the transitions, both the normal and the espilon transitions.
	 * Additionally it shows a few statistics on it (number of states and 
	 * number of transitions) and a list of child automata.</p>
	 * <p>It simply calls the method dump on the tree automata associated
	 * to each element.</p>
	 * @param st the symbol table
	 * @param out the place where to print out
	 * @see xrel.analyzer.TA#dump(PrintStream out)
	 */
	public static void dumpAutomata(SymTable st, PrintStream out) {
		Vector types = st.getVector(SymElementKind.TYPE);
		for (int i=0; i < types.size(); i++) {
			SymElement el = (SymElement)types.elementAt(i);
			TASimple A = el.getElementAutomaton(TABuildPhase.DO_NOTHING,false);
			out.println("Now print the automaton associated to the type " + 
				el.getName() + ":");
			A.dump(out);
			out.println();
		}		

		Vector patterns = st.getVector(SymElementKind.PATTERN);
		for (int i=0; i < patterns.size(); i++) {
			SymElement el = (SymElement)patterns.elementAt(i);
			TASimple A = el.getElementAutomaton(TABuildPhase.DO_NOTHING,false);
			out.println("Now print the automaton associated to the pattern " + 
				el.getName() + ":");
			A.dump(out);
			if (i != patterns.size() - 1) out.println();
		}		
	}

	/**
	 * <p>Builds the automata associated to the elements of the symbol 
	 * table. </p>
	 * <p>Note that each automaton has epsilon transition and unreachable
	 * states removed. </p>
	 * @param st the symbol table.
	 * @param debug required debug info
	 */
	public static void buildAll(SymTable st, boolean debug) {
		Vector types = st.getVector(SymElementKind.TYPE);
		for (int i=0; i < types.size(); i++) {
			SymElement el = (SymElement)types.elementAt(i);
			// the first time is invoked it creates the automaton
			TASimple A = el.getElementAutomaton(TABuildPhase.FULL_BUILD,debug);
		}		

		Vector patterns = st.getVector(SymElementKind.PATTERN);
		for (int i=0; i < patterns.size(); i++) {
			SymElement el = (SymElement)patterns.elementAt(i);
			// the first time is invoked it creates the automaton
			TASimple A = el.getElementAutomaton(TABuildPhase.FULL_BUILD,debug);
		}		
	}
}
