/*
 * PatternMatching.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 29-lug-03
 */
package xrel.analyzer;

import java.util.Vector;
import java.io.PrintStream;

import xrel.XreError;
import xrel.parser.ASTComma;
import xrel.parser.SimpleNode;
import xrel.parser.SymElement;
import xrel.parser.SymElementKind;
import xrel.parser.SymTable;
import xrel.parser.XRelParserTreeConstants;

/**
 * <p>Contains the run-time pattern matching algorithm.
 * The method matchWith calls the algorithm for pattern matching. </p>
 * <p>N.B.:
 * <ul>
 * <li>this is a static class so you will never need to allocate an 
 * object for this class</li>
 * </ul></p>
 * @author Fabrizio Bisi
 */
public class PatternMatching implements SymElementKind {
	private static SymTable symtab = null;
	/**
	 * A return value for the matchWith function: it says what clause 
	 * exactly it matches.
	 */
	private static int matchingClause = -1;
	
	/**
	 * The main method of this class. It tries to match a value passed as
	 * argument against the patterns of the program.
	 * If the match is successful and a typeswitch instruction appears in 
	 * the code the variable matchingClause contains the number of the
	 * clause that matched with the value.
	 * @param value the value to check. The patterns are passed throughout
	 * the symbol table.
	 * @param out where to print out debug information
	 * @param debug if true print out debug information
	 * @return the matching clause or -1
	 */
	public static int matchWith(SymTable st, SimpleNode value, 
																	PrintStream out, boolean debug) {
		symtab = st;
		if (symtab != null) {
			boolean doesMatch;
			Vector patterns = symtab.getVector(PATTERN);
			// get the pattern 0 - that of the import statement
			SymElement pat0 = (SymElement)patterns.elementAt(0);
			TASimple A0 = pat0.getElementAutomaton(TABuildPhase.FULL_BUILD,debug);
			if (debug) out.println("\nExecution of automaton 0:");
			if (A0 != null)
				doesMatch = A0.matchTree(value,debug);
			else {
				XreError.showbug_and_exit("null automaton");
				return -1;
			}
			if (debug) {
				out.println("\nShowing annotations after " +
								   "execution of automaton 0:");
				value.showAnnotations(out,"");
			}
			if (doesMatch) {
				value.bindAll(); // bind variables
//				symtab.showValue(value,out);
				matchingClause = 0;
				if (symtab.originalPatterns > 1) {
					SymElement tswVar = symtab.getTypeSwitchVar();
					if (tswVar != null)  
						value = tswVar.getValue();
					else {
						XreError.showbug_and_exit("typeswitch variable not found");
						return -1;
					}
					if (debug) {
						out.println("\nList of the bound variables " + 
								 "in the import statement:");
						symtab.showBoundVariables(out,true,true);
						out.println();
					}
					doesMatch = false;
					for (int i= 1; i< symtab.originalPatterns; i++) {
						SymElement patI = (SymElement)patterns.elementAt(i);
						TASimple Ai = patI.getElementAutomaton(TABuildPhase.FULL_BUILD,debug);
						if (Ai != null) {
							value.eraseAnnotations();
							if (debug) out.println("Execution of automaton " + i + ":");
							if (Ai.matchTree(value,debug)) {
								if (debug) {
									out.println("\nShowing annotations after " +
									                   "execution of automaton " + i + ":");
									value.showAnnotations(out,"");
								}
								doesMatch = true;
								value.bindAll();
								matchingClause = i;
								if (debug) {
									out.println("\nList of the bound variables " + 
											 "after clause " + i + ":");
									symtab.showBoundVariables(out,true,true);
									out.println();
								}
								break;
							} 
						}
						else {
							XreError.showbug_and_exit("null automaton");
							return -1;
						}
					}
				}
				if (doesMatch) 
					return matchingClause;
				else return -1;
			}
			else {
				matchingClause = -1;
				return matchingClause;
			} 
		}
		else {
			XreError.showbug_and_exit("symbol table uninitialized");
			return -1;
		}
	}
	
//	/**
//	 * Assigns the symbol table. Call this function before to call the main
//	 * method.
//	 * @param st the symbol table pointer.
//	 */
//	public static void setSymbolTable(SymTable st) {
//		symtab = st;
//	}
	
	/**
	 * Evaluates the output value. Call only after executing the 
	 * main algorithm.
	 * @return the output value
	 */
	public static SimpleNode getEvaluation() {
		if (matchingClause > 0) {
			Vector expressions = symtab.getVector(EXPRESSION);
			// however create a comma node
			ASTComma comma = new ASTComma(XRelParserTreeConstants.JJTCOMMA);
			/* then parse the vector of expressions searching for expression with
			 * the right scope */
			for (int i= 0; i < symtab.originalExpressions; i++) {
				SymElement el = (SymElement)expressions.elementAt(i);
				if (matchingClause == el.getScope()) {
					SimpleNode n = el.getNode();
					SimpleNode nCopy = n.jjtmCloneSubtree();
					comma.jjtmAddChildExt(nCopy, comma.jjtGetNumChildren());
				}
			}
			// then if needed remove comma or return null
			if (comma.jjtGetNumChildren() == 0)
				return null;
			else if (comma.jjtGetNumChildren() == 1) {
				SimpleNode child = (SimpleNode)comma.jjtGetChild(0);
				child.jjtSetParent(null);
				return child;
			}
			else return comma;
		}
		else return null;
	}
}
