/*
 * SemanticChecks.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 6-ott-03
 */
package xrel.analyzer;

import java.io.PrintStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

import xrel.XreError;
import xrel.parser.SpecialNames;
import xrel.parser.SymElement;
import xrel.parser.SymElementKind;
import xrel.parser.SymTable;

/**
 * <p>Contains the semantic error checking. Specifically it contains:
 * <dl>
 * <dt>Exhaustivity check</dt>
 * <dd>check that clauses of the typeswitch construct are exhaustive</dd>
 * <dt>Irredundancy check</dt>
 * <dd>check that there are not redundant clauses in the 
 * typeswitch construct</dd>
 * <dt>Ambiguity check</dt>
 * <dd>check that import pattern is not ambiguous and that there are not 
 * ambiguous patterns in clauses of the typeswitch construct. 
 * A pattern is ambiguous when it exists a value that can have multiple 
 * matches</dd>
 * <dt>Output type check</dt>
 * <dd>check that real output types are subtypes of the declared output
 * type</dd>
 * </dl></p>
 * <p>For additional information see the methods of the class. </p>
 * <p>N.B.:
 * <ul>
 * <li>Call these functions only when a typeswitch statement exists!</li>
 * <li>This is a static class, you will never need to allocate an object 
 * of this class</li>
 * </ul></p>
 * @author Fabrizio Bisi
 */
public class SemanticChecks {
	private static int normalizedPatterns = 0;
	
	public static boolean checkTypeSwitch(SymTable st, 
			boolean doIrredundancyChk, boolean doExhaustivityChk,
			boolean doTypeInference, boolean doAmbiguityChk,
			PrintStream out, boolean debug) {
		
		// now take the input type and its automaton (inputType)
		SymElement inputTypeEl = st.getTypeSwitchVar();
		TASimple inputType =
		 inputTypeEl.getElementAutomaton(TABuildPhase.FULL_BUILD,debug);
		if (debug) {
			out.println("\ntype of input value:");
			inputType.dump(out);
		}
		
		/* for each clause compute the union automata and the intersections
		 * of the unions with the input type. This will be useful for
		 * irredundancy check, exhaustivity check and type inference
		 * algorithm */
		// names of the union set of P1 | .. | Pi-1
		HashSet previousNames = new HashSet(); 
		TASimple oldAi = null;
		Vector patterns = st.getVector(SymElementKind.PATTERN);
		for (int i=1; i< st.originalPatterns; i++) {
			if (debug) {
				out.println("\n*********************");
				out.println("* CHECKING CLAUSE " + i + " *");
				out.println("*********************");
			} 
			// get the union of previous clauses P1 | .. | Pi-2
			/* from second iteration update the set of previous clauses
			 * previousClauses = previousClauses | Pi-1 */
			if (oldAi != null) previousNames.add(oldAi.toString());
			TAUnion previousClauses =
			 st.getUnionAutomaton(previousNames,TABuildPhase.FULL_BUILD,debug);
			if (debug) {
				out.println("\nunion of clauses from 1 to " + (i-1) + ":");
				previousClauses.dump(out);
			}
			// get the i-th clause Pi
			SymElement el = (SymElement)patterns.elementAt(i);
			TASimple Ai = el.getElementAutomaton(TABuildPhase.DO_NOTHING,debug);
			if (debug) {
				out.println("\nthe automaton of the clause " + i + " is:");
				Ai.dump(out);
			}
			// update oldAi
			oldAi = Ai;
			// ambiguity check of the current clause
			if (doAmbiguityChk) {
				if(isAmbiguousClause(i,Ai,inputType,st,out,debug)) {
					st.add_warning(st.getClause(i),"clause " + i + " is ambiguous");
					return false;
				}
			}
			// taProduct = inputType x Ai
			TAProduct taProduct =
			 st.getProductAutomaton(inputType,Ai,
									TABuildPhase.STEP_NO_UNMATCHED_STATES,false,debug);
			if (debug) {
				out.println("\nthe product between the input type and the clause " + 
										i + " is:");
				taProduct.dump(out);
			}
			// irredundancy check
			if (doIrredundancyChk) {
				if (!irredundancyCheck(taProduct,previousClauses,st,out,debug)) {
					st.add_type_error(st.getClause(i),"clause " + i + " is redundant");
					return false;
				}
			}
			// type inference
			if (doTypeInference) {
				// I need to simplify before
				// TODO only for type inference?
				taProduct.simplify();
				if (debug) {
					out.println("\nintersection between the input type and " + 
											"the current clause (after simplification)");
					taProduct.dump(out);
				}
				typeInference(taProduct,st,out,debug);
				int exprNum = checkOutputType(i,st,out,debug);
				if (exprNum > 0) {
					st.add_type_error(st.getClause(i),"the expression " + exprNum + 
														" in clause " + i + " is not included in " +
														"output type");
					return false;
				}
			} 
		}
		
		// exhaustivity check
		if (doExhaustivityChk) {
			if (debug) {
				out.println("\n****************************");
				out.println("* GLOBAL TYPESWITCH CHECKS *");
				out.println("****************************");
			} 
			if (oldAi != null) previousNames.add(oldAi.toString());
			if (!exhaustivityCheck(previousNames,inputType,st,out,debug)) {
				st.add_type_error(st.getTypeSwNode(),
								"clauses of the typeswitch construct are not exhaustive");
				return false; 
			}
		}
		return true;
	}
	
	/**
	 * <p>Checks that each clause of the typeswitch construct is not 
	 * redundant. </p>
	 * <p>This means that any new clause Pi matches with at least a value 
	 * of the input type T that is not matched by the previous clauses
	 * P<sub>1</sub>,..,P<sub>i-1</sub>, i.e.:
	 * <blockquote>
	 * for each i, intersection(T,P<sub>i</sub>) \ 
	 * (P<sub>1</sub> | ... | P<sub>i-1</sub>) is not empty
	 * </blockquote>
	 * Note that this check also requires that any clause matches 
	 * at least a value of the input type. </p>
	 * @param st the symbol table
	 * @param out where to print debug information
	 * @param debug heavy debug info
	 * @return true if it's all ok
	 */
	private static boolean irredundancyCheck(
				TAProduct taProduct, TAUnion previousClauses,
				SymTable st, PrintStream out, boolean debug) {
		if (debug) {
			out.println("\nIrredundancy check");
			out.println("------------------");
		} 
		/* For each clause check if it's irredundant
		 * A clause Pi is irredundant w.r.t. input type iff
		 *   ((Pi \ previousClauses) x inputType) is not null
		 * Below we perform the equivalent check that
		 *   ((Pi x inputType) \ previousClauses) is not null
		 * In this way we have two advantages:
		 * 1) we have as intermediate result the computation of 
		 *    (Pi x inputType) that we'll reuse in type inference
		 * 2) we never need to handle ((A\B)xC) that require a
		 *    more complex label-normalization algorithm
		 *    (we could have ~\{a,b} x {a} for example)
		 */
		// taDiff = taProduct \ previousClauses
		TADifference taDiff =
		 st.getDiffAutomaton(taProduct,previousClauses,
							 TABuildPhase.ALL_IN_ONE,debug);
		if (debug) {
			out.println("\nthe difference between resulting product and the " + 
								  "union of the previous clauses is:");
			taDiff.dump(out);
		}
		// check result
		if (taDiff.isNull()) 
			return false; // redundant clause
		else return true; // all ok
	}

	/**
	 * <p>Checks that the clauses of the typeswitch construct are 
	 * exhaustive. </p>
	 * <p>This means that given the input type T and the patterns 
	 * P<sub>1</sub>,..,P<sub>n</sub> of the clauses it must be that T &lt;: 
	 * P<sub>1</sub> | ... | P<sub>n</sub> (the type of the input value 
	 * must be included in the union of the patterns considered as types). </p>
	 * <p><b>Algorithm:</b></p>
	 * <p>We simply verify that T \ (P<sub>1</sub> | .. | P<sub>n</sub>) 
	 * is empty. </p>
	 * @see TADifference
	 * @param st the symbol table
	 * @return true if the clauses are exhaustive
	 */
	private static boolean exhaustivityCheck(
				HashSet previousNames, TASimple inputType, SymTable st, 
				PrintStream out, boolean debug) {
		if (debug) {
			out.println("\nExhaustivity check");
			out.println("------------------");
		} 

		if (debug) {
			out.println("\ntype of input value:");
			inputType.dump(out);
		}

		/* prepare the automaton P1 | ... | Pn
		 * Note that we need only to add the last clause from 
		 * the previous check */
		// get all clauses but the last we found before
		TAUnion allClauses =
			 st.getUnionAutomaton(previousNames,TABuildPhase.ALL_IN_ONE,debug);
//		// take the automaton of the last clause
//		Vector patterns = st.getVector(SymElementKind.PATTERN);
//		SymElement elLast = (SymElement)patterns.elementAt(st.originalPatterns-1);
//		TASimple An = elLast.getElementAutomaton(TABuildPhase.DO_NOTHING,debug);
//		// now add the last clause
//		if (An != null) allClauses.union(An);
		if (debug) {
			out.println("\nunion of clauses from 1 to " + 
									(st.originalPatterns-1) + ":");
			allClauses.dump(out);
		}

		// now get the difference automaton taDiff = inputType \ allClauses 
		TADifference taDiff = st.getDiffAutomaton(inputType,allClauses,
		  TABuildPhase.ALL_IN_ONE,debug); 
	  if (debug) {
		  out.println("\nthe difference between input type and the " + 
								" union of all the clauses is:");
		  taDiff.dump(out);
	  }
	
		/* finally check if it's not empty and if is this the case return
		 * -1 (typeswitch clauses are not exhaustive)
		 */
		boolean isNull = taDiff.isNull();
		if (debug) {
			if (isNull)
				out.println("\nthe difference is empty");
			else
				out.println("\nthe difference is NOT empty");
		}
		return isNull;
	}
	
	public static boolean isAmbiguousImport(SymTable st, 
																					PrintStream out, boolean debug) {
		// take the current clause
		SymElement el0 =
		 (SymElement)st.getVector(SymElementKind.PATTERN).elementAt(0);
		TASimple A0 = el0.getElementAutomaton(TABuildPhase.DO_NOTHING,debug);
		if (debug) {
			out.println("\n***************************");
			out.println("* CHECKING IMPORT PATTERN *");
			out.println("***************************");
		} 
		if (debug) {
			out.println("\nAmbiguity check");
			out.println("---------------");
		} 
		// self product
		TAProduct Ap = st.getProductAutomaton(A0,A0,
				TABuildPhase.STEP_NO_UNREACH_STATES |
				TABuildPhase.STEP_NO_UNMATCHED_STATES,true,debug);
		Ap.update(TABuildPhase.DO_NOTHING);
		if (debug) {
			out.println("\nself product of import pattern");
			Ap.dump(out);
		}
		HashSet emptySet = new HashSet();
		if (recCheckAmbig(emptySet,false,Ap,st,out,debug)) {
			st.add_warning(st.getImportSt(),"import pattern is ambiguous");
			return true;
		} 
		else {
			return false;
		} 
	}
	
	/**
	 * Checks if the patterns of the program are unambiguous.
	 * More specifically it checks if the import pattern is ambiguous for
	 * any input or if any clause is ambiguous with respect to the input type.
	 * It returns -1 when there are no ambiguous patterns, 0 when the 
	 * import pattern is ambiguous or 0..n when the ith clause is ambiguous.
	 * N.B.: the import pattern?
	 * @param st the symbol table
	 * @return the ambiguous pattern or -1
	 */
	private static boolean isAmbiguousClause(int idx, TASimple Ai, 
					TASimple inputType,SymTable st, PrintStream out, boolean debug) {
		if (debug) {
			out.println("\nAmbiguity check");
			out.println("---------------");
		} 
		// self product
		TAProduct Ap = st.getProductAutomaton(Ai,Ai,
				TABuildPhase.STEP_NO_UNREACH_STATES |
				TABuildPhase.STEP_NO_UNMATCHED_STATES,true,debug);
		if (debug) {
			out.println("\nself product of clause " + idx);
			Ap.dump(out);
		}
		TAProduct C = st.getProductAutomaton(inputType,Ap,
			TABuildPhase.STEP_NO_UNREACH_STATES | 
			TABuildPhase.STEP_NO_UNMATCHED_STATES,false,debug);
		if (debug) {
			out.println("\nself product of clause " + idx + 
              " WRT input type " + inputType);
			C.dump(out);
		}
		HashSet emptySet = new HashSet();
		if (recCheckAmbig(emptySet,true,C,st,out,debug)) {
			if (debug) out.println("clause " + idx + " is ambiguous");
			return true;
		} 
		else {
			if (debug) out.println("clause " + idx + " is not ambiguous");
			return false;
		}
	}
	
	private static boolean recCheckAmbig(HashSet alreadyChecked, 
																			boolean wrt, TAProduct Ap, SymTable st, 
																			PrintStream out, boolean debug) {
		// ok I'm checking myself
		alreadyChecked.add(Ap.getName());
		
		// remove transitions with null subautomata
		Ap.simplify();
		
		if (debug) {
			out.println("\n" + Ap + " after simplification:");
			Ap.dump(out);
		}

		// first check that top-level is ambiguous
		if ((!wrt && simpleChkAmb(Ap)) || (wrt && chkAmb(Ap))) 
			return true;
		// else check that also the subautomata are not ambiguous 
		
		// get immediate subautomata
		HashSet subAutomata = Ap.getChildAutomata();
		if (subAutomata == null) {
			XreError.showbug_and_exit("the list of subautomata of " + Ap + " is empty");
			return false;
		}
		
		// then check all subautomata not already checked
		Iterator it = subAutomata.iterator();
		while (it.hasNext()) {
			String saName = (String)it.next();
			if (!alreadyChecked.contains(saName)) {
//				if (debug)
//			 out.println("<<<<<<<<<<<<<<<<<<<< checking ambiguity of "+saName + "..");
				TAProduct sAut = (TAProduct)st.search_automaton(saName);
				if (sAut == null) {
					XreError.showbug_and_exit("can't find automaton " + saName);
					return false;
				}
				if (recCheckAmbig(alreadyChecked, wrt, sAut, st, out, debug))  {
					if (debug)
					 out.println(">>>>>>>>>>>>>>>>>>>> " + saName + "\t is ambiguous");
					return true;
				}
				else if (debug)
				 out.println(">>>>>>>>>>>>>>>>>>>> " + saName + "\t is unambiguous");
			}
		}
		return false; // unambiguous
	}
	
	/* Given a product automaton without useless states
	 * it returns true (is ambiguous) only when it finds <q,q'>
	 * with q != q'
	 */
	private static boolean simpleChkAmb(TAProduct Ap) {
		for (int i=0; i < Ap.states.size(); i++) {
			TAProductState s = (TAProductState)Ap.states.elementAt(i);
			TAState s1 = s.getState1();
			TAState s2 = s.getState2();
			if (s1 != s2)
				return true;
		}
		return false;
	}
	
	/* Given a product automaton without useless states of the
	 * form AxBxB it returns true (is ambiguous) only when it finds
	 * a state <p,<q,r>> such that q != r
	 */
	private static boolean chkAmb(TAProduct Ap) {
		for (int i=0; i < Ap.states.size(); i++) {
			TAProductState s = (TAProductState)Ap.states.elementAt(i);
			TAProductState ss = (TAProductState)s.getState2();
			TAState s1 = ss.getState1();
			TAState s2 = ss.getState2();
			if (s1 != s2) 
				return true;
		}
		return false;
	}
	
	private static int checkOutputType(int clause, SymTable st, 
																				 PrintStream out, boolean debug) {
		if (debug) {
			out.println("\nCheck output type");
			out.println("-----------------");
			out.println("UNIMPLEMENTED!");
		}
//		SymElement outputTypeEl =
//		 st.search(SymElementKind.TYPE,0,SpecialNames.ExportTypeName);
//		if (outputTypeEl != null) {
//			TASimple outputType =
//			 outputTypeEl.getElementAutomaton(TABuildPhase.ALL_IN_ONE,debug);
//			if (outputType == null)
//				XreError.showbug_and_exit("can't build automaton of output type");
//			if (debug) {
//				out.println("\nautomaton of output type:");
//				outputType.dump(out);
//			}
//			int counter = 0;
//			Vector expressions = st.getVector(SymElementKind.EXPRESSION);
//			for (int i=0; i < expressions.size(); i++) {
//				SymElement e = (SymElement)expressions.elementAt(i);
//				if (e.getScope() == clause) {
//					++counter;
//					TASimple eAut = e.getElementAutomaton(TABuildPhase.ALL_IN_ONE,debug);
//					if (debug) {
//						out.println("\nautomaton of expression " + e.getUniqueName() + 
//											  ":");
//						eAut.dump(out);
//					}
//					TADifference taDiff =
//					 st.getDiffAutomaton(eAut,outputType,TABuildPhase.ALL_IN_ONE,debug);
//					if (!taDiff.isNull()) return counter;
//				}
//			}
//		}
		return 0;
	}
	
	/**
	 * Infers the exact type of the variables bound inside the clauses.
	 * This is the type inference algorithm.
	 * @param st the symbol table
	 */
	private static void typeInference(TAProduct taProduct, SymTable st, 
																		PrintStream out, boolean debug) {
		if (debug) {
			out.println("\nType inference");
			out.println("--------------");
		}
		HashSet descAut = taProduct.getDescendantAutomata();
		descAut.add(taProduct.getName());
		Iterator it2 = descAut.iterator();
		while (it2.hasNext()) {
			String autName = (String)it2.next();
			TA aut = st.search_automaton(autName);
			if (aut == null) {
				XreError.showbug_and_exit("can't find " + autName);
			}
			else {
				if (debug) 
					out.println("\nNow processing " + aut);
				// builds the set of the variables bound from the automaton A
				HashSet vars = new HashSet();
				for (int i=0; i < aut.transitions.size(); i++) {
					TATransition t = (TATransition)aut.transitions.elementAt(i); 
					vars.addAll(t.variables);
				}
				// for each variable builds an automaton
				Iterator it = vars.iterator();
				while (it.hasNext()) {
					SymElement el = (SymElement)it.next();
					// inferred types for variables ARE NOT simple but product automata
					if (el.hasAutomaton())
						XreError.showbug_and_exit("attempting to build two times " + 
																			"a var. automaton");
					else {
						TAProduct Av = (TAProduct)aut.clone("#" + el.getAutomatonName());
						filterTransitions(Av,el);
//						if (debug) {
//							out.println("\nautomaton of variable " + el.getUniqueName() + ":");
//							Av.dump(out);
//						}
						Av.force(TABuildPhase.STEP_NO_EPS_TRANS |
									TABuildPhase.STEP_NO_UNREACH_STATES |
								  TABuildPhase.STEP_NO_UNMATCHED_STATES | 
								  TABuildPhase.STEP_COMPLETE_AUT);
						if (debug) {
							out.println("\n(final) automaton of variable " + el.getUniqueName() + 
												  ":");
							Av.dump(out);
						}
						el.setInferredType(Av);
					}
				}
			}
		}
	}
	
	private static void filterTransitions(TA A, SymElement var) {
		// from the last to the first because we can delete the current transition
		for (int i = A.transitions.size()-1; i >= 0; i--) {
			TATransition t = (TATransition)A.transitions.elementAt(i);
			if (! t.variables.contains(var)) {
				TAEpsTransition et = new TAEpsTransition(t.src,t.dst);
				A.epsilonTransitions.add(et);
				A.transitions.remove(i);
			}
		}
	}
}
