/*
 * SymTable.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 12-giu-03
 */
package xrel.parser;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;
import java.io.PrintStream;

import xrel.XreError;
import xrel.analyzer.TA;
import xrel.analyzer.TADifference;
import xrel.analyzer.TASimple;
import xrel.analyzer.TAProduct;
import xrel.analyzer.TAUnion;
//import xrel.xml.Builder;

/**
 * The symbol table contains the types, the patterns, the variables 
 * and the expressions inside the program.
 * @author Fabrizio Bisi
 */
public class SymTable implements XRelParserTreeConstants,SymElementKind, 
                                 SpecialNames {
	/** Contains the types of the program. They're those defined in the
	 * typedef declarations together with the export type.
	 */
	protected Vector types;
	/** Contains the patterns of the program. The first pattern of the
	 * vector is that of the import statement, while the following are
	 * the patterns of the clauses of the typeswitch statement. 
	 */
	protected Vector patterns;
	/** Contains the variables bound inside the patterns of the program. */
	protected Vector variables;
	/** Contains the expressions of the program. Expressions are the 
	 * actual arguments of the printf calls. This vector contains an
	 * ASTStatementList node for each clause of the typeswitch construct, 
	 * so that expressions of the same clause are grouped together. 
	 */
	protected Vector expressions;
	/** Contains the automata inside the program. Inside this vector
	 * you can find an automaton for each pattern, an automaton for 
	 * each local variable for which there is the need to compute the
	 * inferred type and other kinds of automata.
	 */
	protected HashMap automata;
	/** number of program patterns (variable handled at hand from
	 * the extern) */
	public int originalPatterns;
	/** number of program expressions */
	public int originalExpressions;
	/** the name of the typeswitch identifier, if any */
	protected String switchId = null;
	/** a list of clauses - for type error signaling */
	private Vector clauses = null;
	/** the pointer to typeswitch construct */
	private Node typeSwNode = null;
	/** the pointer to import statement */
	private Node importSt = null;
	
	/**
	 * Initializes the symbol table. You don't need to call this
	 * constructor as there is one only symbol table and it's defined
	 * by the parser.
	 */
	public SymTable() {
		types = new Vector();
		patterns = new Vector();
		variables = new Vector();
		expressions = new Vector();
		automata = new HashMap();
		clauses = new Vector();
	}
	
	public void showThings(PrintStream out) {
		out.println();
		out.println(originalPatterns + " original patterns");
		out.println(patterns.size() + " patterns on symbol table");
		out.println(automata.size() + " automata on symbol table");
		dump(out,S_VARIABLES | S_PATTERNS,true,false);
		out.println("\nAutomata:");
		Set automataNames = automata.keySet();
		TreeSet aut = new TreeSet();
		aut.addAll(automataNames);
		java.util.Iterator it = aut.iterator();
		while (it.hasNext()) {
			String s = (String)it.next();
			out.print(s + "\t");
		}
		out.println("\n");
	}
	protected void addClause(Node clause) {
		clauses.add(clause);
	}
	public SimpleNode getClause(int i) {
		return (SimpleNode)clauses.elementAt(i-1);
	}
	protected void setTypeSwNode(Node tswNode) {
		typeSwNode = tswNode;
	}
	public SimpleNode getTypeSwNode() {
		return (SimpleNode)typeSwNode;
	}
	protected void setImportSt(Node impSt) {
		importSt = impSt;
	}
	public SimpleNode getImportSt() {
		return (SimpleNode)importSt;
	}
	/**
	 * Gets the vector of kind k.
	 * @param k the kind of the vector required (TYPE, PATTERN, VARIABLE or
	 * EXPRESSION)
	 */
	public Vector getVector(int k) { 
		Vector v;
		switch(k) {
			case TYPE: v = types; break;
			case PATTERN: v = patterns; break;
			case VARIABLE: v = variables; break;
			case EXPRESSION: v = expressions; break;
			default: 
				XreError.showbug_and_exit("element kind unknown!");
				v = null;
		}
		return v;
	}
	
	/**
	 * Returns the variable element associated to the typeswitch identifier.
	 * @return the element associated to the typeswitch identifier
	 */
	public SymElement getTypeSwitchVar() {
		return search(SymElementKind.VARIABLE,0,switchId);
	}
	/**
	 * Set the typeswitch variable name. This function is used only by the parser.
	 * @param sId
	 */
	public void setSwitchId(String sId) {
		switchId = sId;
	}
	/** Detaches the symbol table elements from the original syntax tree. */
	public void detachPatterns() {
		clean_vector(types);
		clean_vector(patterns);
		//clean_vector(expressions); // useless
		// Warning! I can detach variables as they're inside patterns!
		//clean_vector(variables); 
	}
	
	// at this moment used only by clean_table(). 
	private void clean_vector(Vector v) {
		for (int i = 0; i < v.size(); i++) {
			SymElement el = (SymElement)v.elementAt(i);
			SimpleNode n = el.getNode();
			if (n != null) n.jjtSetParent(null);
		}
	}

	/* --------- */
	/* INSERTION */
	/* --------- */
	
	/**
	 * Puts a type into the symbol table.
	 * @param scope the scope of the element (i.e. its identifier)
	 * @param name the identifier of the element
	 * @param nodeValue the root node associated to the expression that
	 * defines the type
	 */
	public SymElement put_type(int scope, String name, SimpleNode nodeValue) {
		if (name.equals(AnyTypeName))
			add_syntax_error(((SimpleNode)nodeValue.jjtGetParent()),
									"implicit declaration of \"Any\" hidden");
		return put(scope,TYPE,name,nodeValue);
	}
	/**
	 * Puts a variable into the symbol table.
	 * @param scope the scope of the element (i.e. its identifier)
	 * @param name the name of the element
	 * @param nodeId the node of the identifier
	 * @param nodeValue the subtree associated to the element
	 * @return the new symbol element or null in case of error
	 */
	public SymElement put_var(int scope, String name, SimpleNode nodeValue) {
		return put(scope,VARIABLE,name,nodeValue);
	}
	/**
	 * Puts a pattern into the symbol table.
	 * @param scope the scope of the element (i.e. its identifier)
	 * @param name the name of the element
	 * @param nodeValue the subtree associated to the element
	 */
	public void put_pat(int scope, String name, SimpleNode nodeValue) {
		put(scope,PATTERN,name,nodeValue);
	}
	/**
	 * Puts expressions of a code block into the symbol table.
	 * @param scope the scope of the element (i.e. its identifier)
	 * @param name the name of the element
	 * @param nodeValue the ASTStatementList node
	 */
	public void put_expressions(int scope, SimpleNode nodeValue) {
		if (nodeValue.getClass() == ASTStatementList.class) {
			if (nodeValue.jjtGetNumChildren() > 0) 
				for (int i=0; i < nodeValue.jjtGetNumChildren(); i++) {
					SimpleNode n = (SimpleNode)nodeValue.jjtGetChild(i);
					if (n != null) put(scope,EXPRESSION,null,n);
				}
		}
	}

	// now a few support private functions
		
	/* internal put */
	private SymElement put(int scope, int kind, String name, 
									SimpleNode nodeValue) {
		boolean foundError;
		
		/* This function searches first variables, then types, then the rest.
		 * Indeed, between identifiers of the same kind it finds the last 
		 * inserted identifier.
		 * Both these features are used in variable checking: I need
		 * to know if it exists a variable with the same name inside 
		 * the same scoping, a variable with the same name in a previous
		 * scoping or a type name with the same name, in this order. 
		 */
		SymElement oldEl = null;
		if (name != null)
		 oldEl = search(S_VARIABLES|S_TYPES|S_PATTERNS|S_EXPRESSIONS,name);
		if (oldEl == null)
			foundError = false; // new id
		else if (oldEl.getKind() == TYPE)
		{
			/* If there is a type identifier with the same name signal
			 * error, in any case
			 */		
			String ErrMsg = "identifier \"" + name + "\" redeclared"; 
			SimpleNode parent = (SimpleNode)nodeValue.jjtGetParent();
			// signal as error all the binding / all the type declaration
			Token idTk;
			if (parent.getClass() == ASTAs.class) 
				idTk = ((ASTAs)parent).getIdToken();
			else if (parent.getClass() == ASTTypeDecl.class)
				idTk = ((ASTTypeDecl)parent).getIdToken();
			else {
				XreError.showbug_and_exit("unhandled situation in error collecting");
				return null; 
			}
			add_syntax_error(idTk,idTk,ErrMsg);
			foundError = true;
		}
		else if (oldEl.getKind() == VARIABLE) {
			/* if (OldEl.getScope() == scope) then we are inserting
			 * a variable with the same name and the same scope but this
			 * can be good sometimes, for example in "P1 as x | P2 as x".
			 * We check bad situations with checkLinearity
			 */
			if ( (oldEl.getScope() == 0) && (scope > 0) ) {
				String ErrMsg = 
					"identifier \"" + name + "\" hides the previously declared one";
				SimpleNode parent = (SimpleNode)nodeValue.jjtGetParent();
				Token idTk;
				if (parent.getClass() == ASTAs.class) 
					idTk = ((ASTAs)parent).getIdToken();
				else {
					/* variables should be inserted in the symbol table only by
					 * the variable bindings */
					XreError.showbug_and_exit("unhandled situation in error collecting");
					return null; 
				}
				add_warning(idTk,idTk,ErrMsg);
				foundError = false; // a variable hides another one (only a warning)
			}
			else 
				foundError = false; // variable in another context
		}
		else if (oldEl.getKind() == PATTERN || oldEl.getKind() == EXPRESSION) {
			// no error checks on inserting patterns and expressions
			foundError = false; 
		}
		else {
			// What? The kind is?
			XreError.showbug_and_exit("unknown kind of identifier");
			foundError = true; 
		}
		
		if (!foundError) {
			Vector vt = getVector(kind);
			/*
			SimpleNode subtree;
			if (kind == VARIABLE)
				// if it's a variable I copy the subtree, otherwise 
				// I only copy the reference to the subtree
				subtree = nodeValue.jjtmCloneSubtree();
			else 
				subtree = nodeValue;
			*/ 
			SymElement newEl = new SymElement(this,scope, name, kind, nodeValue);
			vt.add(newEl);
			return newEl;
		}
		else return null;
	}
	
	/* -------- */
	/* AUTOMATA */
	/* -------- */
	
	/**
	 * Returns an automaton with a given name. If it doesn't exist
	 * create it.
	 * @param nm the name of the automaton
	 * @return the automaton with the given name
	 */
	public TASimple getSimpleAutomaton(String nm, SymElement el, int BuildPhases,
	                                   boolean debug) {
		TA A = search_automaton(nm);
		if (A==null)
			A = new TASimple(nm,this,debug,System.out,el);
			//automata.put(A.toString(),A); // Warning! it DOESN'T build the automaton
		A.setDebug(debug);
		A.update(BuildPhases);
		return (TASimple)A;
	}

	// Gets A1xA2. It builds the automaton if it doesn't exist
	public TAProduct getProductAutomaton(TA A1, TA A2, int BuildPhases,
	                                     boolean force,boolean debug) {
		String name = TAProduct.getProductName(A1,A2);
		TA A = search_automaton(name);
		if (A==null)
			A = new TAProduct(A1,A2,this,debug,System.out,force);
		A.setDebug(debug);
		BuildPhases |= xrel.analyzer.TABuildPhase.STEP_BUILD;
		A.update(BuildPhases);
		return (TAProduct)A;
	}
	
	public void addAutomaton(TA A) {
		automata.put(A.toString(),A);
	}

	// Gets A1\A2. It builds the automaton if it doesn't exist
	public TADifference getDiffAutomaton(TA A1, TA A2, int BuildPhases,
																			 boolean debug) {
		String name = TADifference.getDiffName(A1,A2);
		TA A = search_automaton(name);
		if (A==null) 
			A = new TADifference(A1,A2,this,debug,System.out);
		A.setDebug(debug);
		BuildPhases |= xrel.analyzer.TABuildPhase.STEP_BUILD;
		A.update(BuildPhases);
		return (TADifference)A;
	}
	
	public TAUnion getUnionAutomaton(HashSet alNames, int BuildPhases,
																	 boolean debug) {
		String name = TAUnion.getUnionName(alNames);
		TA A = search_automaton(name);
		if (A==null) 
			A = new TAUnion(alNames,this,debug,System.out);
		A.setDebug(debug);
		BuildPhases |= xrel.analyzer.TABuildPhase.STEP_BUILD;
		A.update(BuildPhases);
		return (TAUnion)A;
	}
	
	public TA search_automaton(String nm) {
		return (TA)automata.get(nm);
	}
	
	/* --------------- */
	/* MERGE VARIABLES */
	/* --------------- */
	
	/**
	 * <p>Join variables that have the same name and the same scope. </p>
	 * <p>This is a little trick as I need to check syntactic errors and
	 * linearity constraints before to compute a type for the variable. </p> 
	 * <p>This function copies also the subtrees of the variables (it
	 * detaches the variable types from the tree). </p>
	 */
	public void mergeVariables() {
		for (int i=0; i < variables.size(); i++) {
			boolean noFusions = true;
			SymElement v1 = (SymElement)variables.elementAt(i);
			// duplicate v1 and its subtree
			SimpleNode v1Copy = v1.getNode().jjtmCloneSubtree();
			v1.setNode(v1Copy);
			for (int j=i+1; j < variables.size(); j++) {
				SymElement v2 = (SymElement)variables.elementAt(j);
				if (v1.getName().equals(v2.getName()) && 
				    v1.getScope() == v2.getScope()) {
				  // duplicate v2 and its subtree
					SimpleNode v2Copy = v2.getNode().jjtmCloneSubtree();
					v2.setNode(v2Copy);
				  if (noFusions) {
				  	// create "|"
						ASTPar newpar = new ASTPar(JJTPAR);
						newpar.jjtmAddChildExt(v1.getNode(),0);
						v1.setNode(newpar);
						noFusions = false;
				  }
				  SimpleNode parent = v1.getNode();
				  parent.jjtmAddChildExt(v2.getNode(),parent.jjtGetNumChildren());
				  variables.remove(j--);
				}
			}
		}
	}
	
	/* ------ */
	/* SEARCH */
	/* ------ */
	
	/**
	 * Searches an identifier with name "name", scope "scope" and of kind k
	 * (type or pattern or variable or expression).
	 * @param k TYPE or PATTERN or VARIABLE or EXPRESSION
	 * @param scope scoping number: 0 for types and the number of the clause
	 * for variables/patterns/expressions (0 if the variable is bound in the 
	 * import statement or if the pattern is in the import statement)
	 * @param name the name of the identifier to search
	 * @return the searched element or null if not found
	 */
	public SymElement search(int k, int scope, String name) {
		Vector v = getVector(k);

		SymElement result = null;
		for (int i = 0; i < v.size(); i++) {
			SymElement el = (SymElement)v.elementAt(i);
			String n = el.getName(); 
			int s = el.getScope();
			if (n.equals(name) && (s == scope)) {
				result = el;
				break;
			}
		}
		return result;
	}
	
	/**
	 * <P>Searches the last identifier with name "name" inserted in the
	 * symbol table.</P>
	 * <p>This function explicitly searches first variables (if specified)
	 * then patterns (if specified) then everything else.</p>
	 * <P>N.B.: to get the last one I search it from the bottom</P>
	 * @param name the identifier to search
	 * @return the searched element or null
	 */
	public SymElement search(int kinds, String name) {
		SymElement el = null;
		
		// variables have top priority 
		if ((kinds & S_VARIABLES) > 0) {
			el = search(variables,name);
			if (el != null) return el;
		}
		// then types 
		if ((kinds & S_TYPES) > 0) {
			el = search(types,name);
			if (el != null) return el;
		}
		// then everything else
		if ((kinds & S_PATTERNS) > 0) {
			el = search(patterns,name);
			if (el != null) return el;
		}
		if ((kinds & S_EXPRESSIONS) > 0) {
			el = search(expressions,name);
			if (el != null) return el;
		}
		return el;
	}
	
	public SymElement searchPattern(String pStr, int scope) {
		for (int i = 0; i< patterns.size(); i++) {
			SymElement pEl = (SymElement)patterns.elementAt(i);
			if (pEl.getScope() == scope) {
				String currStr = pEl.getNode().getPattern();
				if (pStr.equals(currStr)) return pEl;
			}
		}
		return null;
	}
	
	// now a few private functions
		
	/**
	 * <p> Searches the last identifier inserted in the vector v
	 * with the name n. </p>
	 * <p> This is a low-level private search function.</p>
	 * <p> N.B.: I search from the right to see identifiers
	 * declared in the same context. </p>
	 * @param v vector where to search
	 * @param n the name of the identifier to search
	 * @return the element found or null
	 */
	private SymElement search(Vector v, String n) {
		SymElement result = null;
		for (int i = v.size() - 1; i >= 0; i--) {
			SymElement el = (SymElement)v.elementAt(i);
			if (n.equals(el.getName())) {
				result = el;
				break;
			}
		}
		return result;
	}
	
	/* -------- */
	/* PRINTING */
	/* -------- */

//	/**
//	 * Exports the symbol table to an XML file. Note that at this moment
//	 * we export only the variables.
//	 * @param filename the name to export to
//	 */
//	public void export(String filename) {
//		PrintStream fOut;
//
//		xrel.xml.Builder wr = new xrel.xml.Builder();
//		wr.openDoc(filename,"SymbolTable");
//		wr.openNode("Variables");
//		for (int i = 0; i < variables.size(); i++) {
//			SymElement elem = (SymElement)variables.elementAt(i);
//			wr.openNode("id");
//			wr.addNode("name",elem.getUniqueName());
//			wr.addNode("scope",elem.getScopeStr());
//
//			wr.beginNode("declared_type");
////			fOut = wr.getPrintStream();
////			SimpleNode n = elem.getNode();
////			WriteTokensVisitor tv = new WriteTokensVisitor(fOut);
////			if (n.getFirstToken().kind == XRelParserConstants.SWITCH) {
////				fOut.print("...");
////			}
////			else { n.jjtAccept(tv, null); }
//			wr.endNode("declared_type");
//				
//			wr.beginNode("inferred_type");
//			wr.endNode("inferred_type");
//			wr.closeNode("id");
//		}
//		wr.closeNode("Variables");
//		wr.closeDoc();
//	}
	
	/**
	 * Shows the variables bound in the pattern matching and theirs values.
	 * @param out the print stream where to write out
	 * @param uniqueName if this flag is true variable names are prepended with
	 * the internal unique code
	 * @param internalFormat if this flag is true output is shown in an internal 
	 * format, otherwise in an external (more readable) one.
	 */
	public void showBoundVariables(PrintStream out, 
																 boolean uniqueName, boolean internalFormat) {
		for (int i = 0; i < variables.size(); i++) {
			SymElement el = (SymElement)variables.elementAt(i);
			String name = uniqueName ? el.getUniqueName() : el.getName();
			SimpleNode val = el.getValue();
			if (val != null) {
				out.print(name + " \t=  " );
				if (internalFormat) val.dumpIntFlat(out);
				else val.writePattern(out);
				out.println();
			}
		}
	}
	
	/**
	 * Prints a list of the content of the symbol table to a stream.
	 * @param out the stream where to direct the output
	 */
	public void dump(PrintStream out, int kinds, boolean debug,
	                 boolean internalFormat) {
		if ((kinds & S_TYPES) > 0) {
			out.println("Types:");
			dumpVector(out,types,debug,internalFormat);
		}
		if ((kinds & S_VARIABLES) > 0) {
			out.println("\nVariables:");
			dumpVector(out,variables,debug,internalFormat);
		}
		if ((kinds & S_PATTERNS) > 0) {
			out.println("\nPatterns:");
			dumpVector(out,patterns,debug,internalFormat);
		}
		// now they're lists of statements
		if ((kinds & S_EXPRESSIONS) > 0) {
			out.println("\nExpressions:");
			dumpVector(out,expressions,debug,internalFormat);
		}
	}
	
	/*
	 * Prints the specified vector to the specified stream.
	 * @param out the stream where to direct the output
	 * @param vt the vector to stream out
	 * @param uniqueName specify if it's required the unique name for vars
	 * @param internalFormat specift if to show the internal format
	 */
	private void dumpVector(PrintStream out,Vector vt, 
	                        boolean uniqueName, boolean internalFormat) {
		for (int i = 0; i < vt.size(); i++) {
				SymElement elem = (SymElement)vt.elementAt(i);
				// debug is used to choose between either the name or the unique name
				String name = uniqueName ? elem.getUniqueName() : elem.getName(); 
				SimpleNode node = elem.getNode();
				
				out.print("\t" + name + "\t (" + elem.getScopeStr() + ") \t=  " );
				if (node == null)
					out.println("<null> ");
				else {
					if (internalFormat) elem.getNode().dumpIntFlat(out);
					else elem.getNode().writePattern(out);
					out.println();
				}
		}
	}

	public void dumpTokens(PrintStream out, int kinds, boolean debug) {
		if ((kinds & S_TYPES) > 0) {
			out.println("Types:");
			dumpTokensVector(out,types,debug);
		}
		if ((kinds & S_VARIABLES) > 0) {
			out.println("\nVariables:");
			dumpTokensVector(out,variables,debug);
		}
		if ((kinds & S_PATTERNS) > 0) {
			out.println("\nPatterns:");
			dumpTokensVector(out,patterns,debug);
		}
		if ((kinds & S_EXPRESSIONS) > 0) {
			out.println("\nExpressions:");
			dumpTokensVector(out,expressions,debug);
		}
	}

	private void dumpTokensVector(PrintStream out,Vector vt, boolean debug) {
		for (int i = 0; i < vt.size(); i++) {
			SymElement elem = (SymElement)vt.elementAt(i);
			// debug is used to choose between either the name or the unique name
			String name = debug ? elem.getUniqueName() : elem.getName(); 
			SimpleNode node = elem.getNode();
			
			out.print("\t" + name + "\t (" + elem.getScopeStr() + ") \t=  " );
			if (node == null)
				out.println("<null> ");
			else {
				WriteTokensVisitor wtv = new WriteTokensVisitor(System.out);
				elem.getNode().jjtAccept(wtv,null);
				out.println();
			}
		}
	}

	/* -------------- */
	/* ERROR HANDLING */
	/* -------------- */
	
	protected int ErrorNum = 0;
	protected int WarningNum = 0;
	protected Vector ErrorList = new Vector();
	/**
	 * Maximum number of signaled error, after this limit
	 * error checking stops. 
	 */
	public final int MAX_ERROR_NUM = 60;
	
	/* kind of error/warning */
	private final static String 
		EP_PARSEERROR = "Parsing error",
	 	EP_SYNTAXERROR = "Syntax error",
		EP_TYPEERROR = "Type error",
		EP_WARNING = "Warning";
	
	public void add_parsing_error(SimpleNode n, String msg) {
		addErrListElement(EP_PARSEERROR,msg,n);
	}

	public void add_parsing_error(Token first, Token last, String msg) {
		addErrListElement(EP_PARSEERROR,msg,first,last);
	}
	
	public void add_syntax_error(SimpleNode n, String msg) {
		addErrListElement(EP_SYNTAXERROR,msg,n);
	}
	
	public void add_syntax_error(Token first, Token last, String msg) {
		addErrListElement(EP_SYNTAXERROR,msg,first,last);
	}
	
	public void add_type_error(SimpleNode n, String msg) {
		addErrListElement(EP_TYPEERROR,msg,n);
	}
	
	public void add_type_error(Token first, Token last, String msg) {
		addErrListElement(EP_TYPEERROR,msg,first,last);
	}
	
	public void add_warning(SimpleNode n, String msg) {
		addErrListElement(EP_WARNING,msg,n);
	}
	
	public void add_warning(Token first, Token last, String msg) {
		addErrListElement(EP_WARNING,msg,first,last);
	}
	
	public void checkErrors(SimpleNode rt) {
		ErrorCheckVisitor tv = new ErrorCheckVisitor(this);
		rt.jjtAccept(tv, null);
	}
	
	public void showErrors(PrintStream out, SimpleNode rt) {
		for (int i = 0; i < ErrorList.size(); i++) 
			out.println(ErrorList.elementAt(i).toString());
		out.println("Errors: " + ErrorNum);
		out.println("Warnings: " + WarningNum);
	}
	
	public boolean foundErrors() {
		return ErrorNum != 0;
	}

	public boolean foundWarnings() {
		return WarningNum != 0;
	}

	private void addErrListElement(String kind, String Text, SimpleNode n) {
		addErrListElement(kind,Text,n.getFirstToken(),n.getLastToken());
	}
	
	private void addErrListElement(String kind, String Text, Token first, 
																	Token last) {
		int firstCol = first.beginColumn;
		int firstLine = first.beginLine; 
		int lastCol = last.endColumn;
		int lastLine = last.endLine;
		
		if (kind == EP_WARNING) ++WarningNum;
		else ++ErrorNum;
		if (ErrorNum + WarningNum <= MAX_ERROR_NUM) {
			// search for the position
			int pos = ErrorList.size();
			for (; pos > 0; pos--)
			{
				ErrListElement curr = (ErrListElement)ErrorList.elementAt(pos-1);
				if ((firstLine > curr.firstLine) ||
					 ((firstLine == curr.firstLine) && (firstCol >= curr.firstCol))) {
					break;
				}
			}
			// actually insert in the list of errors
			ErrorList.add(pos, new ErrListElement(kind,Text,
										firstLine,firstCol,lastLine,lastCol));
		}
	}
	// else do nothing, too many errors
}

/**
 * This private class represents a single error line. 
 * @author Fabrizio Bisi
 */
class ErrListElement {
	public String kind;
	public String Text;
	public int firstLine, firstCol, lastLine, lastCol;
	
	public ErrListElement(String knd, String msg, 
										int fromLine, int fromCol,
										int toLine, int  toCol) {
		kind = knd;
		Text = msg;
		firstLine = fromLine;
		firstCol = fromCol;
		lastLine = toLine;
		lastCol = toCol;
	}
	
	public String toString() {
		if (firstLine == lastLine) {
			String chars = lastCol != firstCol ? 
				"characters " + firstCol + "-" + lastCol : 
				"character " + firstCol;
			
			return kind + " " +
				"at line " + firstLine + ", " + chars + ": " + Text;	
		}
		else {
			return kind + " " +
			"from line " + firstLine + ", col. " + firstCol + 
			" to line " + lastLine + ", col. " + lastCol + 
			": " + Text;	
		}
	}
}
